import numpy as np
from sutil.models.Model import Model
from sklearn.neural_network import MLPClassifier
from sklearn.model_selection import GridSearchCV


class NeuralNetworkClassifier(Model):

    def __init__(self, layers, **kwargs):
        self.initialize(layers, kwargs)
    
    def initialize(self, layers, kwargs):
        self.num_labels = layers[-1]
        self.name = kwargs['name'] if 'name' in  kwargs.keys() else 'Neural Network Model'
        self.solver = kwargs['solver'] if 'solver' in  kwargs.keys() else 'adam'
        self.alpha = kwargs['alpha'] if 'alpha' in  kwargs.keys() else 0.001
        self.activation = kwargs['activation'] if 'activation' in  kwargs.keys() else 'relu'
        self.clf = MLPClassifier(solver=self.solver, activation=self.activation, alpha=self.alpha, hidden_layer_sizes=tuple(layers), random_state=1)
    
    def trainModel(self, data):
        print(data.shape)
        self.clf.fit(data.getBiasedX(), data.y.reshape(data.m, ))
    
    def predict(self, X):
        return self.clf.predict(X)
    
    def searchParameters(self, data, param_grid = None, layers_range = (1, 10)):
        layers = []
        for i in range(layers_range[0], layers_range[1]):
            t = (data.n, i, len(data.labels))
            layers.append(t)
        if not param_grid:
            param_grid = [
                    {'activation' : ['identity', 'logistic', 'tanh', 'relu'],
                     'solver' : ['lbfgs', 'sgd', 'adam'],
                     'hidden_layer_sizes': layers}
                    ]
        gs = GridSearchCV(MLPClassifier(), param_grid, cv=3, scoring='accuracy')
        gs.fit(data.X,data.y)
        print("Best parameters set found on development set:")
        print(gs.best_params_)
        best_layer = gs.best_params_['hidden_layer_sizes']
        self.initialize(best_layer, gs.best_params_)


from sutil.text.TextVectorizer import TextVectorizer
from sutil.text.PhraseTokenizer import PhraseTokenizer
from sklearn.feature_extraction.text import TfidfVectorizer
import pandas as pd
import numpy as np

class TFIDFVectorizer(TextVectorizer):

    @classmethod
    def standard(cls):
        return cls({}, PhraseTokenizer())

    def __init__(self, dictionary, tokenizer):
        self.dictionary = dictionary if dictionary else {}
        self.tokenizer = tokenizer
        self.entries = len(dictionary)
        self.vectorizer = TfidfVectorizer()
        self.vectorizer.analyzer = tokenizer.tokenize

    def initialize(self, texts):
        self.index = {}
        self.dictionary = self.colapseTFIDF(texts)
        i = 0
        for fn in self.vectorizer.get_feature_names():
                self.index[fn] = i
                i += 1
        self.entries = len(self.dictionary)

    def colapseTFIDF(self, texts):
        vectors = self.vectorizer.fit_transform(texts)
        dense = vectors.todense()
        denselist = np.array(dense.tolist())
        feature_names = self.vectorizer.get_feature_names()
        averaged = np.zeros(len(denselist[0]))
        for i in range(len(denselist[0])):
            averaged[i] = np.average(denselist[:, i])
        #self.dictionary = pd.DataFrame(denselist, columns=feature_names)
        dictionary = {k: v for k, v in zip(feature_names, averaged)}
        return dictionary

    def getValues(self):
        return self.dictionary.values


# -*- coding: utf-8 -*-
import numpy as np
import pickle
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
from sklearn.decomposition import PCA
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import
import datetime
import random

class Dataset(object):

    def __init__(self, X, y, xlabel='x', ylabel="y", legend=["y=1", "y=0"], title="Graph"):
        self.initialize(X, y, xlabel, ylabel, legend, title)

    @classmethod
    def fromDataFile(cls, filename, delimiter, ysize = 1):
        data = np.loadtxt(filename, delimiter=delimiter)
        X = data[:, 0:-ysize]
        y = data[:, -ysize]
        return cls(X, y)

    @staticmethod
    def load(filename):
        file = open(filename + '.pickle', 'rb')
        return pickle.load(file)

    def initialize(self, X, y, xlabel='x', ylabel="y", legend=None, title="Graph"):
        self.setData(X, y, xlabel, ylabel, legend, title)

    def setData(self, X, y, xlabel='x', ylabel="y", legend=None, title="Graph"):
        self.X = X
        self.y = y
        self.shape = (self.X.shape, self.y.shape)
        self.m, self.n = self.X.shape
        self.size = self.m
        self.y.shape = (self.m, 1)
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.xsize = len(X[0])
        self.ysize = len(y[0])
        self.title = title
        list_set = {*self.y.flatten().tolist()}
        self.labels = (list(list_set))
        if not legend:
            self.legend = []
            for l in self.labels:
                self.legend.append("y=" + str(l))
        else:
            self.legend = legend

    def loadData(self, datafile, delimiter):
        data = np.loadtxt(datafile, delimiter)
        self.X = data[:, 0:-1]
        self.y = data[:, -1]
        self.initialize(self.X, self.y)

    def plotData(self, file=None, d3=False, show_legend = True):
        #The plotted data is assuming 2 dimenssions, check what happened if it's more how to make the projection into 2 dimensions
        classes = self.labels
        if d3:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
        else:
            fig, ax = plt.subplots()
        identifiers = {}
        used = []
        x_graph = self.X
        if len(self.X) > 2:
            pca_analyzer = PCA(n_components=2)
            if d3:
                pca_analyzer = PCA(n_components=3)
            x_graph = pca_analyzer.fit_transform(self.X)

        colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k']
        markers = ['o', 'x', 'v', '^', 's', 'p', 'P', '*', 'h', '1', '2', '3', '4']
        for c in classes:
            if len(used) == len(colors) * len(markers):
                used = []
            assigned = False
            c_index, m_index = 0, 0
            while not assigned:
                c_index, m_index = self.getRandomIdentifier(colors, markers)
                string = colors[c_index] + markers[m_index]
                if string not in used:
                    used.append(string)
                    assigned = True
            index = np.where(self.y == c)
            identifiers[c] = {'color': colors[c_index], 'marker': markers[m_index], "legend": "y=" + str(c)}
            if d3:
                ax.scatter(x_graph[index, 0], x_graph[index, 1], x_graph[index, 2], marker=markers[m_index], c=colors[c_index])
            else:
                ax.scatter(x_graph[index, 0], x_graph[index, 1], marker=markers[m_index], c=colors[c_index])
        ax.set(xlabel=self.xlabel, ylabel=self.ylabel, title=self.title)
        if show_legend:
            ax.legend(self.legend)

        if file:
            plt.savefig(file + '.png')
        plt.show()

    def getRandomIdentifier(self, colors, markers):
        color = random.randint(0, len(colors)-1)
        marker = random.randint(0, len(markers)-1)
        return (color, marker)

    def plotDataRegression(self, file=None, randomize_marker = False):
        fig, ax = self.getPlotRegression(randomize_marker)
        if file:
            fig.savefig(file + ".png")
        plt.show()
        return fig, ax

    def getPlotRegression(self, randomize_marker = False):
        fig, ax = plt.subplots()
        print(self.X.shape, self.y.shape)
        if randomize_marker:
            colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k']
            markers = ['o', 'x', 'v', '^', 's', 'p', 'P', '*', 'h', '1', '2', '3', '4']
            c_index, m_index = self.getRandomIdentifier(colors, markers)
            ax.scatter(self.X, self.y, marker=markers[m_index], c=colors[c_index])
        else:
            ax.scatter(self.X, self.y, marker='x', c='r')
        ax.set(xlabel=self.xlabel, ylabel=self.ylabel, title=self.title)
        ax.legend(self.legend)
        ax.grid()
        return fig, ax

    def normalizeFeatures(self):
        X_norm = self.X;
        mu = np.zeros((self.n, 1))
        sigma = np.zeros((self.n, 1))
        for i in range(self.n):
            mu[i] = np.mean(self.X[:, i])
            sigma[i] = np.std(self.X[:, i])
            if sigma[i] == 0:
                X_norm[:, i] = (self.X[:, i] - mu[i])
            else:
                X_norm[:, i] = (self.X[:, i] - mu[i]) / sigma[i]
        self.X = X_norm
        self.mu = mu
        self.sigma = sigma
        return X_norm, mu, sigma

    def isNormalized(self):
        return (self.mu is not None and self.sigma is not None)

    def normalizeExample(self, example):
        if not self.isNormalized():
            return example
        x_norm = example
        for i in range(len(x_norm)):
            if self.sigma[i] == 0:
                x_norm[i] = (x_norm[i] - self.mu[i]) 
            else:
                x_norm[i] = (x_norm[i] - self.mu[i]) / self.sigma[i]
        return x_norm

    #Returns the test, validation and test datasets
    def split(self, train = 0.8, validation = 0.2):
        #Split train and test
        X_train, X_test, Y_train, Y_test = train_test_split(self.X, self.y, test_size = 1 - train)
        if validation > 0:
            X_train, X_validation, Y_train, Y_validation = train_test_split(X_train, Y_train, test_size = validation)
            return self.clone(X_train, Y_train), self.clone(X_validation, Y_validation), self.clone(X_test, Y_test)
        return self.clone(X_train, Y_train), self.clone(X_test, Y_test)
    
    def save(self, filename = None):
        if not filename:
            generation_time = datetime.datetime.now().strftime('%d_%m_%Y_%H_%M_%S')
            filename = "model_" + generation_time
        filename += ".pickle"
        file = open(filename, 'wb')
        pickle.dump(self, file)

    def sample(self, percentage = None, examples = None):
        if not percentage:
            if not examples:
                return self
            else:
                percentage = max(0, examples/self.m)
        X_train, X_test, Y_train, Y_test = train_test_split(self.X, self.y, test_size = percentage)
        return Dataset(X_test, Y_test)

    def getShape(self):
        return self.X.shape, self.y.shape

    def getBiasedX(self):
        return np.hstack((np.ones((len(self.X), 1)), self.X))

    def clone(self, X, y):
        return  Dataset(X, y, self.xlabel, self.ylabel, self.legend, self.title)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Friday Jan 17 18:28:06 2020
pre processor class to clean text
@author: mcampos
"""
import re
from nltk.corpus import stopwords
from nltk.stem import PorterStemmer
from nltk.stem import WordNetLemmatizer
from sutil.text.Num2Words import Num2Words
from sutil.text.StringJanitor import StringJanitor

class PreProcessor:

    @classmethod
    def standard(cls):
        configuration = [("case", "lower"),
                         ("denoise", "spanish")]
        return cls(configuration)

    def __init__(self, configurations):
        self.actions = []
        m2m = {"case": "caseNormalization",
               "denoise": "removeNoise",
               "stopwords": "stopWordsRemoval",
               "stem": "stem",
               "lemmatize": "lemmatize",
               "normalize": "normalize",
               "num2words": "convertNum2Words",
               "callable":"callable"}
        for entry in configurations:
        	self.actions.append((m2m[entry[0]], entry[1]))
        self.lemmatizer = WordNetLemmatizer()
        self.janitor = StringJanitor.spanish()
        self.stemmer = PorterStemmer()
        self.num2words = Num2Words()

    def preProcess(self, string):
        result = str(string)
        for a in self.actions:
            method = getattr(self, a[0])
            result = method(a[1], result)
        return result

    def stopWordsRemoval(self, idiom, string):
        sw = set(stopwords.words(idiom))
        words = string.split(" ")
        cleaned_words = [w for w in words if w not in sw]
        return " ".join(cleaned_words)

    def removeNoise(self, idiom, string):
        cleaned_string = self.janitor.clean(string)
        return cleaned_string.replace("_", " ")

    def caseNormalization(self, type, string):
        if type == "lower":
            return string.lower()

    def stem(self, idiom, string):
        raw_words = string.split(" ")
        stemmed_words = [self.stemmer.stem(word=word) for word in raw_words]
        return " ".join(stemmed_words)

    def lemmatize(self, idiom, string):
        words = string.split(" ")
        lemmatized_words = [self.lemmatizer.lemmatize(word=word, pos='v') for word in words]
        return " ".join(lemmatized_words)

    def convertNum2Words(self, idiom, string):
        new_string = self.num2words.replace(string)
        return new_string

    def callable(self, function, string):
        return function(string)

    def batchPreProcess(self, texts):
        cleaned = []
        for t in texts:
            cleaned.append(self.preProcess(t))
        return cleaned

    #Simple text normalization using regular expressions
    def normalize(self, patterns, string):
        normalized = string
        for p in patterns:
            normalized = re.sub(p[0], p[1], normalized)
        return normalized


from sutil.text.PhraseTokenizer import PhraseTokenizer

class GramTokenizer(PhraseTokenizer):

    def __init__(self, gram_length=3, space_char='_'):
        self.gram_length = gram_length
        self.space_char = space_char

    def getTokens(self, text, delimiter = " "):
        """
        This method extract the grams of the string
        """
        #copy = super(GramTokenizer, self).clean_string(text)
        copy = text.replace(" ", self.space_char)

        if len(copy) <= self.gram_length:
            grams = [copy.ljust(self.gram_length, self.space_char)]
        else:
            grams_number = len(copy) - self.gram_length + 1
            grams = [copy[i:i+self.gram_length] for i in range(grams_number)]
        return grams


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr  4 16:14:03 2019
@author: kampamocha
"""
import re
from textwrap import wrap

class Num2Words:
    """Convert string containing a number to its text form"""

    def __init__(self, separator=(".", "punto"), prefixes={"-": "menos", "+": "mas"}, chunks=(0,2)):
        """Class constructor - receives parameters for number and text format"""
        self.separator = separator
        self.prefixes = prefixes
        self.chunks = chunks

    def to_string(self, number):
        return str(number)

    def to_number(self, text_number):
        if text_number.isdigit():
            return int(text_number)
        else:
            return float(text_number)

    def find(self, text_number):
        numeric_const_pattern = "[" + "".join(self.prefixes.keys()) + "]? "
        numeric_const_pattern += "(?: (?: \d* \\"
        numeric_const_pattern += self.separator[0]
        numeric_const_pattern += " \d+ ) | (?: \d+ \\"
        numeric_const_pattern += self.separator[0]
        numeric_const_pattern += "? ) )"

        rx = re.compile(numeric_const_pattern, re.VERBOSE)
        return rx.finditer(text_number)
        #return rx.findall(text_number)

    def replace(self, text):
        new = text[:]
        for m in self.find(text):
            space_before = space_after = ""
            pos_before, pos_after = m.span()[0] - 1, m.span()[1] + 1
            if pos_before >= 0:
                if not text[pos_before].isspace():
                    space_before = " "
            if pos_after < len(text):
                if not text[pos_after].isspace():   # Add space if next character is not
                    space_after = " "
            new = new.replace(m.group(), space_before + self.convert(m.group()) + space_after, 1)
        return new

    def convert(self, text_number):
        words = []
        text_to_process = text_number
        if text_number[0] in self.prefixes:
            words.append(self.prefixes[text_number[0]])
            text_to_process = text_to_process[1:]

        segments = text_to_process.split(self.separator[0])

        for i in range(len(segments)):
            ch = self.chunks[i] if i < len(self.chunks) else self.chunks[-1]
            groups = [segments[i]] if not ch else wrap(segments[i], ch)
            for number in groups:
                words.append(self.__int2spanish(number))
            if i < len(segments)-1:
                words.append(self.separator[1])

        return " ".join([w for w in words if w])

    def __int2spanish(self, text_number):
        # Validate input is only digits
        if not text_number.isdigit():
            return text_number

        if text_number[0] == '0':
            number_letters = 'cero'
            rest = self.__int2spanish(text_number[1:])
            if rest:
                number_letters += (" " + rest)
            return number_letters

        number = int(text_number)

        indicator = [("",""),("mil","mil"),("millón","millones"),("mil","mil"),("billón","billones")]
        integer = number
        counter = 0
        number_letters = ""

        while integer > 0:
            a = integer % 1000
            if counter == 0:
                in_letters = self.__convert_spanish(a,1).strip()
            else:
                in_letters = self.__convert_spanish(a,0).strip()
            if a==0:
                number_letters = in_letters + " " + number_letters
            elif a==1:
                if counter in (1,3):
                    number_letters = indicator[counter][0] + " " + number_letters
                else:
                    number_letters = in_letters + " " + indicator[counter][0] + " " + number_letters
            else:
                number_letters = in_letters + " " + indicator[counter][1] + " " + number_letters
            number_letters = number_letters.strip()
            counter += 1
            integer = int(integer / 1000)

        return number_letters

    def __convert_spanish(self, number, sw):
        hundreds_list = ["", ("cien", "ciento"), "doscientos", "trescientos", "cuatrocientos", "quinientos",
                         "seiscientos", "setecientos", "ochocientos", "novecientos"]
        tens_list = ["", ("diez", "once", "doce", "trece", "catorce", "quince", "dieciseis", "diecisiete",
                          "dieciocho", "diecinueve"), ("veinte", "veinti"), ("treinta", "treinta y"),
                         ("cuarenta", "cuarenta y"), ("cincuenta", "cincuenta y"), ("sesenta", "sesenta y"),
                         ("setenta", "setenta y"), ("ochenta", "ochenta y"), ("noventa", "noventa y") ]
        unit_list = ["", ("un", "uno"), "dos", "tres", "cuatro", "cinco", "seis", "siete", "ocho", "nueve"]
        hundreds = int(number / 100)
        tens = int((number - (hundreds * 100)) / 10)
        units = int(number - (hundreds * 100 + tens * 10))

        units_text = ""

        # Validate hundreds
        hundreds_text = hundreds_list[hundreds]
        if hundreds == 1:
            if (tens + units) != 0:
                hundreds_text = hundreds_text[1]
            else:
                hundreds_text = hundreds_text[0]

        # Validate tens
        tens_text = tens_list[tens]
        if tens == 1:
            tens_text = tens_text[units]
        elif tens > 1:
            if units != 0:
                tens_text = tens_text[1]
            else:
                tens_text = tens_text[0]

        #Validate units
        if tens != 1:
            units_text = unit_list[units]
            if units == 1:
                units_text = units_text[sw]

        frmt = "%s %s%s" if tens_text == 'veinti' else "%s %s %s"
        return frmt %(hundreds_text, tens_text, units_text)



#        r"""
#            [-+]? # optional sign
#            (?:
#                (?: \d* \. \d+ ) # .1 .12 .123 etc 9.1 etc 98.1 etc
#                |
#                (?: \d+ \.? ) # 1. 12. 123. etc 1 12 123 etc
#            )
#            # followed by optional exponent part if desired
#            # (?: [Ee] [+-]? \d+ ) ?
#            """



import numpy as np
from sutil.text.TextVectorizer import TextVectorizer

class OneHotVectorizer(TextVectorizer):

    def initialize(self, texts):
        self.index = {}
        for t in texts:
            tokens = self.tokenizer.tokenize(t)
            for token in tokens:
                self.addTermToDictionary(token)

    def addTermToDictionary(self, term):
        if not term in self.dictionary:
            self.dictionary[term] = 1
            self.entries += 1
            self.index[term] = len(self.index)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 13 22:28:06 2019
string_utils
@author: kampamocha
"""
import string
import unicodedata

class StringJanitor:
    
    def __init__(self, common_chars=string.ascii_lowercase + string.digits, space_char='_', additional_chars=''):
        """Class constructor - receives characters to validate"""        
        self.common_chars = common_chars
        self.space_char = space_char
        self.additional_chars = additional_chars
        self.valid_chars = self.common_chars + self.space_char + self.additional_chars
           
    def clean(self, text):
        """Common method to sanitize the text"""
        sanitized = text.lower()
        sanitized = sanitized.replace(self.space_char,' ')
        sanitized = self.space_char.join(sanitized.split())
        chars = [c if c in self.valid_chars else unicodedata.normalize('NFD', c)[0] for c in sanitized]
        sanitized = ''.join([c for c in chars if c in self.valid_chars])
        return sanitized 

    # Class methods for initialization
    @classmethod
    def spanish(cls):
        """String janitor for spanish"""
        common_chars = string.ascii_lowercase + string.digits
        space_char = '_'
        additional_chars = 'ñ'
        return cls(common_chars=common_chars, space_char=space_char, additional_chars=additional_chars)

    @classmethod
    def portuguese(cls):
        """String janitor for portuguese"""
        common_chars = string.ascii_lowercase + string.digits
        space_char = '_'
        additional_chars = 'ç'
        return cls(common_chars=common_chars, space_char=space_char, additional_chars=additional_chars)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 29 22:47:44 2020

@author: mcampos
"""

# -*- coding: utf-8 -*-
import numpy as np

class MultiClassF1:
    def __init__(self, predictions, values, classes = None):
        self.predictions = predictions
        self.values = values
        self.classes = classes if classes is not None else np.unique(values)
        self.scores = {}
        self.calculateF1()
        
    
    def calculateF1(self):
        
        for c in self.classes:
            tp = np.sum((self.predictions == c) & (self.values == c))
            tn = np.sum((self.predictions != c) & (self.values != c))
            fp = np.sum((self.predictions == c) & (self.values != c))
            fn = np.sum((self.predictions != c) & (self.values == c))
            precision = tp/(tp + fp)
            recall = tp/(tp + fn)
            f1 = 2 * (precision*recall)/(precision+recall)
            self.scores[c] = {'tp':tp, 'tn':tn, 'fp':fp, 'fn':fn, 'precision': precision, 'recall': recall, 'f1': f1}
        return self.scores
    
    def average(self):
        precision_total, recall_total, f1_total = 0, 0, 0
        num_classes = 0
        for c in self.classes:
            if c in self.scores:
                num_classes += 1
                precision_total += self.scores[c]['precision']
                recall_total += self.scores[c]['recall']
                f1_total += self.scores[c]['f1']
                
        return {'avg_precision': precision_total/num_classes, 
                'avg_recall': recall_total/num_classes, 
                'avg_f1': f1_total/num_classes}


# -*- coding: utf-8 -*-

class Experiment:

    def __init__(self, data, models = None, train = 0.8, validation=0.2, k_folds = 3):
        self.models = models if models is not None else {}
        self.models_index = {}
        self.train_p = train
        self.validation_p = validation
        self.train, self.validation, self.test = None, None, None
        self.splitData(data)

    def splitData(self, data):
        if self.train is None:
            if self.validation_p > 0:
                self.train, self.validation, self.test = data.split(self.train_p, self.validation_p)
            else:
                if self.train_p >= 1:
                    self.train = data
                else:
                    self.train, self.test = data.split(self.train_p, self.validation_p)
        return True

    def addModel(self, model, modelClass = None, name = None):
        index = len(self.models)
        name = name if name else "Modelo " + str(len(self.models) + 1)
        self.models[name] = model
        self.models_index[name] = index

    def model(self, i = None, name = None):
        if i:
            return self.models[self.models_index[i]]
        else:
            if name:
                return self.models[name]
        return None

    def run(self, plot = False, train = False):
        for model_name, model in self.models.items():
            print("Training model " + model_name)
            model.trainModel(self.train)
            if train:
                print("*" * 30)
                print("Training score")
                model.score(self.train.X, self.train.y)
            if self.validation is not None:
                print("*" * 30)
                print("Validation score")
                model.score(self.validation.X, self.validation.y)
            if self.test is not None:
                print("*" * 30)
                print("Test score")
                model.score(self.test.X, self.test.y)
            if plot:
                model.roc.plot()


from sutil.models.Model import Model
from tensorflow.keras.layers import Dense, Input, LSTM, Embedding, Dropout
from tensorflow.keras.layers import GlobalMaxPool1D
from tensorflow.keras.models import Model as KerasModel
from sklearn import preprocessing
import matplotlib.pyplot as plt
import numpy as np

class LSTMModel(Model):

    @classmethod
    def fromDataset(cls, data, input_size, output_size, embedding_size):
        lb = preprocessing.LabelBinarizer()
        lb.fit(data.y)
        return cls(input_size, 'LSTM Model', 20000, embedding_size, output_size, lb)

    def __init__(self, maxlen, name = 'LSTM Model', max_features = 20000, embed_size = 128, number_of_classes = 1, binarizer = None):
        self.name = name
        self.binarizer = binarizer
        self.model = self.defineModel(maxlen, max_features, embed_size, number_of_classes)
        self.model.summary()

    def defineModel(self, maxlen, max_features = 20000, embed_size = 128, number_of_classes = 1):
        #Input layer
        inp = Input(shape=(maxlen, )) 
        x = Embedding(max_features, embed_size)(inp)
        #LSTM layer
        x = LSTM(60, return_sequences=True,name='lstm_layer')(x)
        x = GlobalMaxPool1D()(x)
        x = Dropout(0.1)(x)
        x = Dense(50, activation="relu")(x)
        x = Dropout(0.1)(x)
        x = Dense(number_of_classes, activation="softmax")(x)
        model = KerasModel(inputs=inp, outputs=x)
        model.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])
        return model

    def trainModel(self, data, batch_size = 32, epochs = 2):
        binary_y = self.binarizer.transform(data.y)
        self.model.fit(data.getBiasedX(),binary_y, batch_size=batch_size, epochs=epochs, validation_split=0.1)
        
    def predict(self, X):
        return self.model.predict(X)
    
    def score(self, X, y, bias = True):
        if bias:
            X = np.hstack((np.ones((len(X), 1)), X))
        predictions = self.predict(X)
        self.evaluation = self.model.evaluate(X, y, batch_size=128)
        try:
            self.classification_report = classification_report(y, predictions)
            print(self.classification_report)
            self.accuracy = accuracy_score(y, predictions, normalize=False)
            self.recall = recall_score(y, predictions)
            self.precission = precision_score(y, predictions)
            self.f1 = f1_score(y, predictions)
            #self.roc = ModelROC(self, X, y, self.name)
        except:
            print("An exception occurred")

    def plotHistory(self):
        history = self.model.history
        acc = history.history['acc']
        val_acc = history.history['val_acc']
        loss = history.history['loss']
        val_loss = history.history['val_loss']
        # Get number of epochs\n",
        epochs = range(len(acc))

        # Plot training and validation accuracy per epoch\n",
        plt.plot(epochs, acc)
        plt.plot(epochs, val_acc)
        plt.title('Training and validation accuracy')
        plt.figure()

        # Plot training and validation loss per epoch\n",
        plt.plot(epochs, loss)
        plt.plot(epochs, val_loss)
        plt.title('Training and validation loss')
        plt.show()


# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
from sklearn.metrics import auc
from sklearn.metrics import roc_curve

class ModelROC:
    def __init__(self, model, X_test, y_test, legend = 'model'):
        self.model = model
        y_pred = self.getPredictions(X_test)
        fpr, tpr, treshold = roc_curve(y_test, y_pred)
        self.fpr = fpr
        self.tpr = tpr
        self.treshold = treshold
        self.auc = auc(fpr, tpr)
        self.legend = legend
        print("Area under the curve " + self.legend + " " + str(self.auc))
    
    def getPredictions(self, X_test):
        return self.model.predict(X_test)
    
    def plotData(self):
        plt.plot([0, 1], [0, 1], 'k--')
        plt.plot(self.fpr, self.tpr, label= self.legend + ' (area = {:.3f})'.format(self.auc))
        plt.xlabel('False positive rate')
        plt.ylabel('True positive rate')
        plt.title('ROC curve')
        plt.legend(loc='best')
        
    def plot(self):
        plt.figure(1)
        self.plotData()
        plt.show()
        
    def zoom(self, zoom_x, zoom_y):
        plt.figure(2)
        plt.xlim(zoom_x[0], zoom_x[1])
        plt.ylim(zoom_y[0], zoom_y[1])
        self.plotData()
        plt.show()



import numpy as np

class TextVectorizer():

    def __init__(self, dictionary, tokenizer):
        self.dictionary = dictionary if dictionary else {}
        self.tokenizer = tokenizer
        self.entries = len(dictionary)
        self.index = {}
        for k, v in self.dictionary.items():
            if not k in self.index.keys():
                self.index[k] = len(self.index)

    def initialize(self, texts):
        self.index = {}
        for t in texts:
            for token in self.tokenizer.tokenize(t):
                if token in self.dictionary.keys():
                    self.dictionary[token] += 1 
                else:
                    self.dictionary[token] = 1
                    self.index[token] = len(self.index)
        # Encode the reciprocal
        for k in self.dictionary.keys():
            self.dictionary[k] = 1 / (self.dictionary[k] + 1)
        self.entries = len(self.dictionary)

    def encodePhrase(self, phrase):
        vector = np.zeros(len(self.dictionary))
        for t in self.tokenizer.tokenize(phrase):
            if t in self.dictionary.keys():
                vector[self.index[t]] = self.dictionary[t]
        return vector

    def decodeVector(self, vector):
        inv_dict = {v: k for k, v in self.index.items()}
        out = ""
        for i in range(len(vector)):
            if vector[i] != 0:
                out += str(inv_dict[i]) + ","
        return out[: -1]

    def textToMatrix(self, texts):
        print("Texts: " + str(len(texts)))
        print("Dictionary: " + str(len(self.dictionary)))
        matrix = np.zeros((len(texts), self.entries))
        if len(self.dictionary) < 1:
            return np.array([0])
        i = 0
        for t in texts:
            matrix[i:] = self.encodePhrase(t)
            i += 1
            if i % 500 == 0:
                print(i)
        return matrix


# -*- coding: utf-8 -*-
from sutil.base.Dataset import Dataset
from sutil.text.PreProcessor import PreProcessor
from sutil.text.PhraseTokenizer import PhraseTokenizer
from sutil.text.OneHotVectorizer import OneHotVectorizer
import pandas as pd

class TextDataset(Dataset):

    @classmethod
    def standard(cls, filename, delimiter):
        preprocessor = PreProcessor.standard()
        tokenizer = PhraseTokenizer()
        vectorizer = OneHotVectorizer({}, tokenizer)
        df = pd.read_csv(filename)
        X = df[df.columns[0]]
        y = df[df.columns[-1]].values
        return cls(X, y, vectorizer, preprocessor, False)

    @classmethod
    def setvectorizer(cls, filename, vectorizer = None, preprocessor = None, normalize_features = False, **kwargs):
        df = pd.read_csv(filename)
        X = df[df.columns[0]]
        y = df[df.columns[-1]].values
        return cls(X, y, vectorizer, preprocessor)

    def __init__(self, X, y, vectorizer = None, preprocessor = None, normalize_features = False, **kwargs):
        self.normalize_features = normalize_features
        if vectorizer:
            self.vectorizer = vectorizer 
        else: 
            self.vectorizer = OneHotVectorizer()
        if preprocessor:
            self.preprocessor = preprocessor 
        else: 
            self.preprocessor = PreProcessor.standard()
        #print(X)
        self.initialize(X, y)

    def initialize(self, texts, y, **kwargs):
        self.texts = texts
        processed = self.preprocessor.batchPreProcess(texts)
        self.vectorizer.initialize(processed)
        X = self.vectorizer.textToMatrix(processed)
        #print(X)
        #print(y)
        self.setData(X, y)
        if self.normalize_features:
            self.X, self.mu, self.sigma = self.normalizeFeatures()

    def encodePhrase(self, phrase):
        phrase = self.preprocessor.preProcess(phrase)
        #return self.normalizeExample(self.vectorizer.encodePhrase(phrase))
        if self.normalize_features:
            return self.normalizeExample(self.vectorizer.encodePhrase(phrase))
        return self.vectorizer.encodePhrase(phrase)

    def decodeVector(self, vector):
        return self.vectorizer.decodeVector(vector)



import numpy as np
import scipy.optimize as op
import scipy as scipy
from sutil.base.Dataset import Dataset
from sutil.models.Model import Model

class RegularizedLogisticRegression(Model):

    def __init__(self, theta, alpha, l, **kwargs):
        self.theta = theta
        self.alpha = alpha
        self.l = l
        self.name = kwargs['name'] if 'name' in  kwargs.keys() else 'Logistic Regression'
        
    
    @classmethod
    def fromDataset(cls, data, alpha = 0.1, l=0.1):
        theta = np.random.random(data.X.shape[1] + 1)
        return cls(data, theta, alpha, l)
    
    @classmethod
    def fromDataFile(cls, datafile, delimeter, alpha=0.1, l=0.1):
        data = Dataset.fromDataFile(datafile, delimeter)
        theta = np.random.random(data.X.shape[1] + 1)
        return cls(data, theta, alpha, l)

    #m denotes the number of examples
    #gradient indicates the gradient matrix
    #regularization is the regularization parameter in order to prevent the over fitting
    #cost is the cotst of the logistic regression funciton
    def getCostAndGradient(self, data, theta):
        gradient = np.zeros(np.size(theta))
        m = data.m
        h_theta = self.computePredictions(data, theta).reshape(m, 1)
        regularization = (self.l * np.sum(theta[1 :]**2))/(2*m)
        
        CK = 0.000009
        h_theta[h_theta == 0] = CK
        h_theta[h_theta == 1] = 1 - CK
        
        cost_m = data.y * np.log(h_theta) + (1 - data.y) * np.log(1 - h_theta)
        
        cost = -1 * np.sum(cost_m)/m + regularization

        #We calculate the gradient of the first regression parameter
        differences = h_theta - data.y
        X = data.getBiasedX()
        gradient[0] = np.sum(differences * X[:, 0].reshape(m, 1))/m
        for j in range(1, len(theta)):
            gradient[j] = np.sum(differences * X[:, j].reshape(m, 1))/m - self.l/m * theta[j]
        return cost, gradient

    def getCostAndGradient2(self, data, theta):
        m = data.m
        gradient = np.zeros(np.size(theta))
        h_theta = self.computePredictions(data, theta).reshape(m, 1)
        regularization = (self.l * np.sum(theta[1 :]**2))/(2*m)
        cost_m = self.train.y * np.log(h_theta) + (1 - data.y) * np.log(1 - h_theta)
        cost = -1 * np.sum(cost_m)/m + regularization

        #We calculate the gradient of the first regression parameter
        differences = h_theta - data.y
        X = data.getBiasedX()
        gradient[0] = np.sum(differences * X[:, 0].reshape(m, 1))/m
        for j in range(1, len(theta)):
            gradient[j] = np.sum(differences * X[:, j].reshape(m, 1))/m - self.l/m * theta[j]
        return cost, gradient
    
    @staticmethod
    def getCost(theta, x, y, l=0):
        m = len(x)
        sigmoid = lambda z: 1/ (1 + np.exp(-1 * z))
        #sigmoid = lambda z: scipy.special.expit(z)
        h_theta = sigmoid(np.matmul(x, theta)).reshape(m, 1)
        regularization = (l * np.sum(theta[1 :]**2))/(2*m)
        cost_m = y * np.log(h_theta) + (1 - y) * np.log(1 - h_theta)
        cost = -1 * np.sum(cost_m)/m + regularization
        return cost
    
    @staticmethod
    def getGradient(theta, x, y, l=0):
        gradient = np.zeros(np.size(theta))
        m = len(x)
        sigmoid = lambda z: 1/ (1 + np.exp(-1 * z))
        #input("Wait a second...")
        h_theta = sigmoid(np.matmul(x, theta)).reshape(m, 1)
        differences = h_theta - y
        gradient[0] = np.sum(differences * x[:, 0].reshape(m, 1))/m
        for j in range(1, len(theta)):
            gradient[j] = np.sum(differences * x[:, j].reshape(m, 1))/m - l/m * theta[j]
        return gradient
                   
    def sigmoid(self, z):
        return scipy.special.expit(z)
        
    def gradienDescent(self, data, iterations):
        theta = self.theta
        cost = []
        gradient = []
        for i in range(iterations):
          costj, gradientj = self.getCostAndGradient(data, theta)
          cost.append(costj)
          gradient.append(gradientj)
          theta[0] -= self.alpha * gradientj[0]
          for j in range(1, len(theta)):
              theta[j] -= self.alpha * (gradientj[j] + (self.l/data.m) * theta[j])
        self.theta = theta
        return self.theta, cost, gradient

    def optimizedGradientDescent(self, data):
        initial_theta = np.zeros(data.n + 1)
        result = op.minimize(fun = RegularizedLogisticRegression.getCost, 
                             x0 = initial_theta, 
                             args = (data.getBiasedX(), data.y),
                             method = 'TNC',
                             jac = RegularizedLogisticRegression.getGradient)
        return result

    def computePredictions(self, data, theta):
        return self.sigmoid(np.matmul(data.getBiasedX(), theta))

    def evaluateHypotesis(self, xi, theta):
        return self.sigmoid(np.matmul(xi, theta))

    def makePrediction(self, x):
        evaluation = self.evaluateHypotesis(x, self.theta)
        pred = np.vectorize(lambda x: 1 if x >= 0.5 else 0)
        return pred(evaluation)
    
    def predict(self, X):
        return self.makePrediction(X)
    
    def trainModel(self, data, epochs = 100):
        result = self.optimizedGradientDescent(data)
        self.theta = result.x
        return self.theta


import string
import unicodedata

class PhraseTokenizer():

    def __init__(self, delimiter = None):
        self.delimiter = delimiter if delimiter else " "

    def getTokens(self, text, delimiter = " "):
        return text.split(delimiter)

    def tokenize(self, text):
        return self.getTokens(text)



# -*- coding: utf-8 -*-
import sutil.base.Dataset as Dataset
from sklearn.metrics import accuracy_score
from sklearn.metrics import recall_score
from sklearn.metrics import f1_score
from sklearn.metrics import precision_score
from sklearn.metrics import classification_report
from sutil.metrics.ModelROC import ModelROC
import numpy as np

class Model(object):
    
    def __init__(self, name = 'Model'):
        self.name = name
    
    def fit(self, X = None, y = None):
        data = Dataset(X, y)
        self.trainModel(data, 100)   
        
    def predict(self, X):
        return X
    
    def score(self, X, y, bias = True):
        if bias:
            X = np.hstack((np.ones((len(X), 1)), X))
        predictions = self.predict(X)
        try:
            self.classification_report = classification_report(y, predictions)
            print(self.classification_report)
            self.accuracy = accuracy_score(y, predictions, normalize=False)
            self.recall = recall_score(y, predictions)
            self.precission = precision_score(y, predictions)
            self.f1 = f1_score(y, predictions)
            #self.roc = ModelROC(self, X, y, self.name)
        except:
            print("An exception occurred")



import numpy as np
import scipy.optimize as op
import sutil.base.Dataset as Dataset
from sutil.models.Model import Model

class RegularizedLinearRegression(Model):

    def __init__(self, theta, alpha, l, **kwargs):
        self.theta = theta
        self.alpha = alpha
        self.l = l
    
    @classmethod
    def fromDataset(cls, data, alpha = 0.1, l=0.1):
        theta = np.rand(data.y.shape)
        return cls(data, theta, alpha, l)
    
    @classmethod
    def fromDataFile(cls, datafile, delimeter, alpha=0.1, l=0.1):
        data = Dataset.fromDataFile(datafile, delimeter)
        theta = np.rand(data.y.shape)
        return cls(data, theta, alpha, l)

    #m denotes the number of examples
    #gradient indicates the gradient matrix
    #regularization is the regularization parameter in order to prevent the over fitting
    #cost is the cotst of the logistic regression funciton
    def getCostAndGradient(self, data, theta):
        X = data.getBiasedX()
        m = data.m
        gradient = np.zeros(np.size(theta))
        h_theta = self.evaluateHypothesis(X, theta).reshape(m, 1)
        differences = h_theta - data.y
        cost_m = differences **2
        regularization = (self.l * np.sum(theta[1 :]**2))/(2*m)
        cost = np.sum(cost_m)/(2*m) + regularization        
        gradient[0] = np.sum(differences * X[:, 0].reshape(len(X), 1))/m
        #We calculate the gradient for the rest of the parameters
        for j in range(1, len(theta)):
            gradient[j] = np.sum(differences * X[:, j].reshape(len(X), 1))/m + self.l/m * theta[j]
        return cost, gradient
    
    @staticmethod
    def getCost(theta, x, y, l=0):
        m = len(x)
        differences = np.matmul(x, theta).reshape(m, 1) - y
        cost_m = differences **2
        regularization = (l * np.sum(theta[1 :]**2))/(2*m)
        cost = np.sum(cost_m)/(2*m) + regularization
        return cost
    
    @staticmethod
    def getGradient(theta, x, y, l=0):
        gradient = np.zeros(np.size(theta))
        m = len(x)
        differences = np.matmul(x, theta).reshape(m, 1) - y
        gradient[0] = np.sum(differences * x[:, 0].reshape(m, 1))/m
        #We calculate the gradient for the rest of the parameters
        for j in range(1, len(theta)):
            gradient[j] = np.sum(differences * x[:, j].reshape(m, 1))/m + (l/m) * theta[j]
        return gradient

    def gradienDescent(self, data, iterations):
        theta = self.theta
        cost = []
        gradient = []
        print("****************************Computing gradient descent***************************")
        print(iterations)
        for i in range(iterations):
          costj, gradientj = self.getCostAndGradient(data, theta)
          cost.append(costj)
          gradient.append(gradientj)
          for j in range(len(theta)):
              theta[j] -= self.alpha * gradientj[j]
        self.theta = theta
        return self.theta, cost, gradient
    
    @staticmethod
    def normalEquation(data):
        X = data.getBiasedX()
        theta = np.matmul(np.linalg.pinv(np.matmul(X.T, X)), np.matmul(X.T, data.y))
        return theta

    def optimizedGradientDescent(self, data):
        initial_theta = np.zeros(data.n + 1)
        result = op.minimize(fun = RegularizedLinearRegression.getCost, 
                             x0 = initial_theta, 
                             args = (data.getBiasedX(), data.y),
                             method = 'TNC',
                             jac = RegularizedLinearRegression.getGradient)
        return result

    def evaluateHypothesis(self, xi, theta):
        return np.matmul(xi, theta)
    
    def makePrediction(self, x):
        return self.evaluateHypothesis(x, self.theta)
    
    def predict(self, X):
        return self.makePrediction(X)
    
    def trainModel(self, data, iterations = 100):
        return self.gradienDescent(data, iterations)


# -*- coding: utf-8 -*-
from sutil.models.Model import Model

class SklearnModel(Model):
    
    def __init__(self, name, external):
        self.name = name
        self.external = external
    
    def trainModel(self, data):
        #Normalize and bias
        self.external.fit(data.getBiasedX(), data.y)
    
    def predict(self, X):
        return self.external.predict(X)
